/* Image filter library for sqcam driver by philippe Corbes (04/02/2007)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This source code is based heavily on the gspca webcam driver which is
 * written by Peter Pregler, Scott J. Bertin and Johannes Erdfelt
 *
 * Portions of this code were also copied from usbvideo.c
 *
 * The bayer decoding functions are copied from libgphoto2
 *
 */

#include "image_filter.h"
#include "gamma_table.h"

/* Image parameters adjustement */
static int Gamma       = FILTER_DEFAUL_GAMMA;
static int OffsetRed   = FILTER_DEFAUL_OFFSET;
static int OffsetGreen = FILTER_DEFAUL_OFFSET;
static int OffsetBlue  = FILTER_DEFAUL_OFFSET;
static int GainRed     = FILTER_DEFAUL_GAIN;
static int GainGreen   = FILTER_DEFAUL_GAIN;
static int GainBlue    = FILTER_DEFAUL_GAIN;

/* /sys/module/sqcam/parameters */
module_param(Gamma,       int, 0644);
module_param(OffsetRed,   int, 0644);
module_param(OffsetGreen, int, 0644);
module_param(OffsetBlue,  int, 0644);
module_param(GainRed,     int, 0644);
module_param(GainGreen,   int, 0644);
module_param(GainBlue,    int, 0644);

MODULE_PARM_DESC(Gamma,       "Gamma setting range 0 to 6 3-> Gamma=1");
MODULE_PARM_DESC(OffsetRed,   "OffsetRed setting range -128 to 128");
MODULE_PARM_DESC(OffsetGreen, "OffGreen setting range -128 to 128");
MODULE_PARM_DESC(OffsetBlue,  "OffBlue setting range -128 to 128");
MODULE_PARM_DESC(GainRed,     "Gain Red setting range 0 to 512 /256 ");
MODULE_PARM_DESC(GainGreen,   "Gain Green setting range 0 to 512 /256 ");
MODULE_PARM_DESC(GainBlue,    "Gain Blue setting range 0 to 512 /256 ");

#define CLIP(color) (unsigned char)(((color)>0xFF)?0xff:(((color)<0)?0:(color)))

/****************************************************************************
*  Image filter
***************************************************************************/
static void
image_filter_set(struct image_filter *filter)
{
	filter->enabled     = 0;
	filter->change      = FILTER_UPDATE_GAMMA;
	filter->Gamma       = Gamma;
	filter->OffsetRed   = OffsetRed;
	filter->OffsetGreen = OffsetGreen;
	filter->OffsetBlue  = OffsetBlue;
	filter->GainRed     = GainRed;
	filter->GainGreen   = GainGreen;
	filter->GainBlue    = GainBlue;
	image_filter_update(filter);
}


static void
image_filter_update(struct image_filter *filter)
{	
    unsigned char *red   = filter->red;
    unsigned char *green = filter->green;
    unsigned char *blue  = filter->blue;
    int i;

    if (filter->change) {
		if (filter->change == FILTER_UPDATE_GAMMA) {
		    /* Gamma setting change compute all case */
		    memcpy(red,   GammaTable[filter->Gamma], 256);
		    memcpy(green, GammaTable[filter->Gamma], 256);
		    memcpy(blue,  GammaTable[filter->Gamma], 256);
		    for (i = 0; i < 256; i++) {
				red[i]   = CLIP(((red[i]   + filter->OffsetRed)   * filter->GainRed) >> 8);
				green[i] = CLIP(((green[i] + filter->OffsetGreen) * filter->GainGreen) >> 8);
				blue[i]  = CLIP(((blue[i]  + filter->OffsetBlue)  * filter->GainBlue) >> 8);
		    }
		    // filter->change &= ~FILTER_UPDATE_GAMMA;   // Geändert 
		    filter->change = FILTER_UPDATED;
		}
		if (filter->change == FILTER_UPDATE_RED) {
		    /* Red setting change compute Red Value */
		    memcpy(red, GammaTable[filter->Gamma], 256);
		    for (i = 0; i < 256; i++) {
				red[i] = CLIP(((red[i] + filter->OffsetRed) * filter->GainRed) >> 8);
		    }
		    filter->change &= ~FILTER_UPDATE_RED;
		}
		if (filter->change == FILTER_UPDATE_GREEN) {
		    /* Green setting change compute Green Value */
		    memcpy(green, GammaTable[filter->Gamma], 256);
		    for (i = 0; i < 256; i++) {
				green[i] = CLIP(((green[i] + filter->OffsetGreen) * filter->GainGreen) >> 8);
		    }
		    filter->change &= ~FILTER_UPDATE_GREEN;
		}
		if (filter->change == FILTER_UPDATE_BLUE) {
		    /* Blue setting change compute Blue Value */
		    memcpy(blue, GammaTable[filter->Gamma], 256);
		    for (i = 0; i < 256; i++) {
				blue[i] = CLIP(((blue[i] + filter->OffsetBlue) * filter->GainBlue) >> 8);
		    }
		    filter->change &= ~FILTER_UPDATE_BLUE;
		}
		/* Finaly check if the filter need to be applied */
		filter->enabled = (  (filter->Gamma       != FILTER_DEFAUL_GAMMA)
						  || (filter->OffsetRed   != FILTER_DEFAUL_OFFSET)
						  || (filter->OffsetGreen != FILTER_DEFAUL_OFFSET)
						  || (filter->OffsetBlue  != FILTER_DEFAUL_OFFSET)
						  || (filter->GainRed     != FILTER_DEFAUL_GAIN)
						  || (filter->GainGreen   != FILTER_DEFAUL_GAIN)
						  || (filter->GainBlue    != FILTER_DEFAUL_GAIN)
						  );
    }
}
  
static void
image_filter_apply (struct image_filter *filter, unsigned char *image, unsigned int size)
{
	int x;

	for (x = 0; x < (size * 3); x += 3) {
		image[x + 0] = filter->red  [image[x + 0]];
		image[x + 1] = filter->green[image[x + 1]];
		image[x + 2] = filter->blue [image[x + 2]];
	}
}

